﻿using System;
using System.Globalization;
using System.Xml;
using System.Text;
using MathEx.ExternalBallistic.JBM;
using MathEx.ExternalBallistic.Units;

namespace MathEx.ExternalBallistic
{
    public static class ShotInfoController
    {
        public static class Calculation
        {
            public static bool calculateZero(ShotInfo zeroInfo)
            {
                if (!canCalculate(zeroInfo))
                    return false;


                Atmo atm = null;

                if (zeroInfo.Atmosphere != null)
                    atm = new Atmo(zeroInfo.Atmosphere.Temperature.Get(Temperature.Unit.Fahrenheit),
                                     zeroInfo.Atmosphere.Pressure.Get(Pressure.Unit.InchHg),
                                     zeroInfo.Atmosphere.Humidity,
                                     zeroInfo.Atmosphere.Altitude.Get(Distance.Unit.Foot));
                else
                    atm = new Atmo();

                Bullet bullet = new Bullet(zeroInfo.Ammo.BallisticCoefficient, zeroInfo.Ammo.Table);
                Rifle rifle = new Rifle(zeroInfo.Ammo.MuzzleVelocity.Get(Velocity.Unit.FeetPerSecond), zeroInfo.SightHeight.Get(Distance.Unit.Foot));

                Wind wind = null;
                if (zeroInfo.Wind != null)
                    wind = new Wind(zeroInfo.Wind.Speed.Get(Velocity.Unit.FeetPerSecond), zeroInfo.Wind.Direction.Get(Angle.Unit.Radian));

                Shot shot = new Shot(zeroInfo.ShotAngle != null ? zeroInfo.ShotAngle.Get(Angle.Unit.Radian) : 0, zeroInfo.CantAngle != null ? zeroInfo.CantAngle.Get(Angle.Unit.Radian) : 0, 0);

                double zero = zeroInfo.ZeroDistance.Get(Distance.Unit.Foot);

                JbmCalculator.calculate(bullet, rifle, atm, shot, wind, zero, zeroInfo.NearZero, 0, 1, true, false);

                zeroInfo.ElevationAngle = new Angle(shot.Elevation, Angle.Unit.Radian);

                return true;
            }

            public static BallisticInfoCollection calculateShot(ShotInfo shotInfo)
            {
                if (!canCalculate(shotInfo))
                    return null;

                Atmo atm = null;

                if (shotInfo.Atmosphere != null)
                    atm = new Atmo(shotInfo.Atmosphere.Temperature.Get(Temperature.Unit.Fahrenheit),
                                     shotInfo.Atmosphere.Pressure.Get(Pressure.Unit.InchHg),
                                     shotInfo.Atmosphere.Humidity,
                                     shotInfo.Atmosphere.Altitude.Get(Distance.Unit.Foot));
                else
                    atm = new Atmo();

                Bullet bullet = new Bullet(shotInfo.Ammo.BallisticCoefficient, shotInfo.Ammo.Table);
                Rifle rifle = new Rifle(shotInfo.Ammo.MuzzleVelocity.Get(Velocity.Unit.FeetPerSecond), shotInfo.SightHeight.Get(Distance.Unit.Foot));

                Wind wind = null;
                if (shotInfo.Wind != null)
                    wind = new Wind(shotInfo.Wind.Speed.Get(Velocity.Unit.FeetPerSecond), shotInfo.Wind.Direction.Get(Angle.Unit.Radian));

                Shot shot = new Shot(shotInfo.ShotAngle != null ? shotInfo.ShotAngle.Get(Angle.Unit.Radian) : 0, shotInfo.CantAngle != null ? shotInfo.CantAngle.Get(Angle.Unit.Radian) : 0, shotInfo.ElevationAngle != null ? shotInfo.ElevationAngle.Get(Angle.Unit.Radian) : 0);

                RangeData[] ballistic = JbmCalculator.calculate(bullet, rifle, atm, shot, wind, 0, false, shotInfo.MaxDistance.Get(Distance.Unit.Foot), shotInfo.Step.Get(Distance.Unit.Foot), false, true);

                if (ballistic == null)
                    return null;

                BallisticInfoCollection result = new BallisticInfoCollection();
                result.Reserve(ballistic.Length);

                double bulletWeightGr = shotInfo.Ammo.BulletWeight.Get(Weight.Unit.Grain);
                double stablityCoefficient = 0;
                if (shotInfo.DriftInfo != null)
                    stablityCoefficient = DriftInfoController.Calculation.calculateStabilityCoefficient(shotInfo.Ammo, shotInfo.DriftInfo, shotInfo.Atmosphere);

                for (int i = 0; i < ballistic.Length && ballistic[i] != null; i++)
                {
                    double v = ballistic[i].Velocity;
                    double wi = ballistic[i].Windage;
                    double si = 0;
                    if (shotInfo.DriftInfo != null)
                        si = DriftInfoController.Calculation.calculateDrift(shotInfo.DriftInfo, stablityCoefficient, ballistic[i].Time).Get(Distance.Unit.Foot);
                    wi += si;


                    Distance range = new Distance(ballistic[i].Range, Distance.Unit.Foot);
                    Distance path = new Distance(ballistic[i].Drop, Distance.Unit.Foot);
                    Angle hold = new Angle(Math.Atan(ballistic[i].Drop / ballistic[i].Range), Angle.Unit.Radian);
                    Velocity velocity = new Velocity(v, Velocity.Unit.FeetPerSecond);
                    Distance windage = new Distance(wi, Distance.Unit.Foot);
                    Angle windageCorrection = new Angle(Math.Atan(wi / ballistic[i].Range), Angle.Unit.Radian);
                    Weight ogv = new Weight(Math.Pow(v, 3) * Math.Pow(bulletWeightGr, 2) * 1.5e-12, Weight.Unit.Pound);
                    Energy energy = new Energy(bulletWeightGr * v * v / 450400, Energy.Unit.FootPounds);
                    BallisticInfo ballisticInfo = new BallisticInfo(range, path, hold, TimeSpan.FromSeconds(ballistic[i].Time), windage, windageCorrection, velocity, ballistic[i].Mach, energy, ogv);
                    result.Add(ballisticInfo);
                }
                return result;
            }

            public static bool canCalculate(ShotInfo shot)
            {
                return shot != null &&
                       shot.Ammo != null &&
                       shot.SightHeight != null;
            }
        }

        public static class Serialization
        {
            public static void Write(XmlDocument doc, XmlNode parent, ShotInfo info)
            {
                XmlNode node = doc.CreateElement("shot-info");
                XmlAttribute attr;
                if (info.Name != null)
                {
                    attr = doc.CreateAttribute("name");
                    attr.Value = info.Name;
                    node.Attributes.Append(attr);
                }

                if (info.ShotAngle != null)
                {
                    attr = doc.CreateAttribute("shot-angle");
                    attr.Value = info.ShotAngle.ToString(info.ShotAngle.SetUnit, false, 8);
                    node.Attributes.Append(attr);
                }

                if (info.CantAngle != null)
                {
                    attr = doc.CreateAttribute("cant-angle");
                    attr.Value = info.CantAngle.ToString(info.CantAngle.SetUnit, false, 8);
                    node.Attributes.Append(attr);
                }

                if (info.ElevationAngle != null)
                {
                    attr = doc.CreateAttribute("elevation-angle");
                    attr.Value = info.ElevationAngle.ToString(info.ElevationAngle.SetUnit, false, 8);
                    node.Attributes.Append(attr);
                }

                if (info.MaxDistance != null)
                {
                    attr = doc.CreateAttribute("max-distance");
                    attr.Value = info.MaxDistance.ToString(info.MaxDistance.SetUnit, false, 8);
                    node.Attributes.Append(attr);
                }

                if (info.SightHeight != null)
                {
                    attr = doc.CreateAttribute("sight-height");
                    attr.Value = info.SightHeight.ToString(info.SightHeight.SetUnit, false, 8);
                    node.Attributes.Append(attr);
                }

                if (info.ZeroDistance != null)
                {
                    attr = doc.CreateAttribute("zero-distance");
                    attr.Value = info.ZeroDistance.ToString(info.ZeroDistance.SetUnit, false, 8);
                    node.Attributes.Append(attr);
                }

                if (info.Step != null)
                {
                    attr = doc.CreateAttribute("step");
                    attr.Value = info.Step.ToString(info.Step.SetUnit, false, 8);
                    node.Attributes.Append(attr);
                }

                if (info.Ammo != null)
                    AmmoInfoController.Serialization.Write(doc, node, info.Ammo);

                if (info.Atmosphere != null)
                    AtmosphereInfoController.Serialization.Write(doc, node, info.Atmosphere);

                if (info.Wind != null)
                    WindInfoController.Serialization.Write(doc, node, info.Wind);

                if (info.DriftInfo != null)
                    DriftInfoController.Serialization.Write(doc, node, info.DriftInfo);

                if (parent == null)
                    doc.AppendChild(node);
                else
                    parent.AppendChild(node);
            }

            public static ShotInfo Read(XmlNode node)
            {
                ShotInfo shot = new ShotInfo();

                foreach (XmlAttribute attr in node.Attributes)
                {
                    if (attr.Name == "max-distance")
                    {
                        Distance v;
                        Distance.Unit u;
                        if (Distance.TryParse(attr.Value, out v, out u))
                            shot.MaxDistance = v;
                    }
                    else if (attr.Name == "step")
                    {
                        Distance v;
                        Distance.Unit u;
                        if (Distance.TryParse(attr.Value, out v, out u))
                            shot.Step = v;
                    }
                    else if (attr.Name == "sight-height")
                    {
                        Distance v;
                        Distance.Unit u;
                        if (Distance.TryParse(attr.Value, out v, out u))
                            shot.SightHeight = v;
                    }
                    else if (attr.Name == "zero-distance")
                    {
                        Distance v;
                        Distance.Unit u;
                        if (Distance.TryParse(attr.Value, out v, out u))
                            shot.ZeroDistance = v;
                    }
                    else if (attr.Name == "shot-angle")
                    {
                        Angle v;
                        Angle.Unit u;
                        if (Angle.TryParse(attr.Value, out v, out u))
                            shot.ShotAngle = v;
                    }
                    else if (attr.Name == "cant-angle")
                    {
                        Angle v;
                        Angle.Unit u;
                        if (Angle.TryParse(attr.Value, out v, out u))
                            shot.CantAngle = v;
                    }
                    else if (attr.Name == "elevation-angle")
                    {
                        Angle v;
                        Angle.Unit u;
                        if (Angle.TryParse(attr.Value, out v, out u))
                            shot.ElevationAngle = v;
                    }
                    else if (attr.Name == "name")
                    {
                        shot.Name = attr.Value;
                    }
                }

                shot.Ammo = AmmoInfoController.Serialization.FindAndRead(node);
                shot.Atmosphere = AtmosphereInfoController.Serialization.FindAndRead(node);
                shot.Wind = WindInfoController.Serialization.FindAndRead(node);
                shot.DriftInfo = DriftInfoController.Serialization.FindAndRead(node);
                return shot;
            }

            public static void WriteFile(string name, ShotInfo info)
            {
                XmlDocument doc = new XmlDocument();
                Write(doc, null, info);

                XmlWriterSettings settings = new XmlWriterSettings();
                settings.ConformanceLevel = ConformanceLevel.Document;
                settings.Encoding = Encoding.UTF8;
                settings.Indent = true;
                settings.IndentChars = (" ");
                settings.OmitXmlDeclaration = true;

                XmlWriter writer;
                writer = XmlWriter.Create(name, settings);
                doc.Save(writer);
                writer.Close();

            }

            public static ShotInfo ReadFile(string name)
            {
                XmlReaderSettings settings;
                settings = new XmlReaderSettings();
                settings.ConformanceLevel = ConformanceLevel.Document;
                settings.IgnoreWhitespace = true;
                settings.IgnoreComments = true;

                XmlReader reader = XmlReader.Create(name);
                XmlDocument doc = new XmlDocument();
                doc.Load(reader);
                reader.Close();

                return Read(doc.DocumentElement);
            }
        }
    }
}