﻿using System;
using System.Globalization;

namespace MathEx.ExternalBallistic.Units
{
    public class Angle
    {
        //angle in radians
        private double mValue;
        private static Angle mConvertor = new Angle(0, Unit.Radian);
        private static object mMutex = new object();

        public enum Unit
        {
            Radian,
            Degree,
            Moa,
            Mil,
            MilDot,
            AMil
        }

        public Angle(double value, Unit unit)
        {
            Set(value, unit);
        }

        public double Get(Unit unit)
        {
            switch (unit)
            {
            case    Unit.Radian:
                    return mValue;
            case    Unit.Degree:
                    return mValue * 180 / Math.PI;
            case    Unit.MilDot:
                    return mValue * 3200 / Math.PI;
            case    Unit.AMil:
                    return mValue * 3000 / Math.PI;
            case    Unit.Moa:
                    return mValue * 60 * 180 / Math.PI;
            case    Unit.Mil:
                    return mValue * 1000;
            }
            throw new ArgumentException("Unknown unit");
        }

        private Unit mSetUnit;

        public Unit SetUnit
        {
            get
            {
                return mSetUnit;
            }
        }

        protected void Set(double value, Unit unit)
        {
            mSetUnit = unit;
            switch (unit)
            {
            case    Unit.Radian:
                    mValue = value;
                    break;
            case    Unit.Degree:
                    mValue = value * Math.PI / 180;
                    break;
            case    Unit.MilDot:
                    mValue = value * Math.PI / 3200;
                    break;
            case    Unit.AMil:
                    mValue = value * Math.PI / 3000;
                    break;
            case    Unit.Moa:
                    mValue = value / 60 * Math.PI / 180;
                    break;
            case    Unit.Mil:
                    mValue = value / 1000;
                    break;
            }
        }

        static public double Convert(double value, Unit from, Unit to)
        {
            lock (mMutex)
            {
                mConvertor.Set(value, from);
                return mConvertor.Get(to);
            }
        }

        static public string UnitToName(Unit unit)
        {
            switch (unit)
            {
            case    Unit.Radian:
                    return "rad";
            case    Unit.Degree:
                    return "°";
            case    Unit.Moa:
                    return "moa";
            case    Unit.Mil:
                    return "mil";
            case    Unit.AMil:
                    return "am";
            case    Unit.MilDot:
                    return "mdot";
            }
            throw new ArgumentException("Unknown unit");
        }

        static public Unit NameToUnit(string name)
        {
            if (name == "rad")
                return Unit.Radian;
            else if (name == "deg" || name == "°")
                return Unit.Degree;
            else if (name == "moa")
                return Unit.Moa;
            else if (name == "mil")
                return Unit.Mil;
            else if (name == "am")
                return Unit.AMil;
            else if (name == "mdot")
                return Unit.MilDot;
            throw new ArgumentException("Unknown unit");
        }

        static public int DefaultDisplayPrecision(Unit unit)
        {
            switch (unit)
            {
            case    Unit.Radian:
                    return 6;
            case    Unit.Degree:
                    return 1;
            case    Unit.Moa:
                    return 1;
            case    Unit.Mil:
                    return 1;
            case    Unit.MilDot:
                    return 1;
            case    Unit.AMil:
                    return 1;
            }
            throw new ArgumentException("Unknown unit");
        }

        static public Unit DefaultUnit
        {
            get
            {
                return Unit.Radian;
            }
        }

        override public string ToString()
        {
            return ToString(DefaultUnit, true, DefaultDisplayPrecision(DefaultUnit));
        }

        public string ToString(Unit unit)
        {
            return ToString(unit, true, DefaultDisplayPrecision(unit));
        }

        public string ToString(Unit unit, bool groupThousands, int precision)
        {
            NumberFormatInfo fi = new NumberFormatInfo();
            fi.NumberDecimalDigits = precision;
            fi.CurrencyDecimalSeparator = ".";
            fi.NumberGroupSeparator = ",";
            if (groupThousands)
                fi.NumberGroupSizes = new int[] {3, 0};
            else
                fi.NumberGroupSizes = new int[] {0};
            fi.NumberNegativePattern = 1;
            return Get(unit).ToString("N", fi) + UnitToName(unit);
        }

        static public bool TryParse(string text, out Angle value, out Unit unit)
        {
            value = null;
            unit = DefaultUnit;

            //split value to text and unit name
            string number = null;
            string name = null;
            for (int i = text.Length - 1; i >= 0; i--)
            {
                if (Char.IsDigit(text[i]))
                {
                    name = text.Substring(i + 1);
                    number = text.Substring(0, i + 1);
                    break;
                }
            }

            if (name == null || name.Length == 0)
                return false;

            try
            {
                unit = NameToUnit(name);
            }
            catch (Exception )
            {
                return false;
            }

            NumberFormatInfo fi = new NumberFormatInfo();
            fi.CurrencyDecimalSeparator = ".";
            fi.NumberGroupSeparator = ",";
            fi.NumberNegativePattern = 1;

            double v;
            if (!Double.TryParse(number, out v))
                return false;

            value = new Angle(v, unit);
            return true;
        }

        public Angle ToUnit(Unit unit)
        {
            return new Angle(Get(unit), unit);
        }
    };
}