﻿using System;
using System.Globalization;

namespace MathEx.ExternalBallistic.Units
{
    public class Pressure
    {
        private double mValue;            //value in mmHg
        static private Pressure mConvertor = new Pressure(0, Unit.MmHg);
        static private object mMutex = new object();

        public enum Unit
        {
            MmHg,
            InchHg,
            Bar,
        }

        public Pressure(double value, Unit unit)
        {
            Set(value, unit);
        }

        private Unit mSetUnit;

        public Unit SetUnit
        {
            get
            {
                return mSetUnit;
            }
        }

        protected void Set(double value, Unit unit)
        {
            mSetUnit = unit;

            switch (unit)
            {
            case    Unit.MmHg:
                    mValue = value;
                    break;
            case    Unit.InchHg:
                    mValue = value * 25.4;
                    break;
            case    Unit.Bar:
                    mValue = value * 750.061683;
                    break;
            }
        }

        public double Get(Unit unit)
        {
            switch (unit)
            {
            case    Unit.MmHg:
                    return mValue;
            case    Unit.InchHg:
                    return mValue / 25.4;
            case    Unit.Bar:
                    return mValue / 750.061683;
            }
            throw new ArgumentException("Unknown unit");
        }

        static public double Convert(double value, Unit from, Unit to)
        {
            lock (mMutex)
            {
                mConvertor.Set(value, from);
                return mConvertor.Get(to);
            }
        }

        static public string UnitToName(Unit unit)
        {
            switch (unit)
            {
            case    Unit.MmHg:
                    return "mmHg";
            case    Unit.InchHg:
                    return "inHg";
            case    Unit.Bar:
                    return "bar";
            }
            throw new ArgumentException("Unknown unit");
        }

        static public Unit NameToUnit(string name)
        {
            if (name == "mmHg")
                return Unit.MmHg;
            else if (name == "inHg")
                return Unit.InchHg;
            else if (name == "bar")
                return Unit.Bar;
            throw new ArgumentException("Unknown unit");
        }

        static public int DefaultDisplayPrecision(Unit unit)
        {
            switch (unit)
            {
            case    Unit.MmHg:
                    return 0;
            case    Unit.InchHg:
                    return 2;
            case    Unit.Bar:
                    return 4;
            }
            throw new ArgumentException("Unknown unit");
        }

        static public Unit DefaultUnit
        {
            get
            {
                return Unit.MmHg;
            }
        }

        override public string ToString()
        {
            return ToString(DefaultUnit, true, DefaultDisplayPrecision(DefaultUnit));
        }

        public string ToString(Unit unit)
        {
            return ToString(unit, true, DefaultDisplayPrecision(unit));
        }

        public string ToString(Unit unit, bool groupThousands, int precision)
        {
            NumberFormatInfo fi = new NumberFormatInfo();
            fi.NumberDecimalDigits = precision;
            fi.CurrencyDecimalSeparator = ".";
            fi.NumberGroupSeparator = ",";
            if (groupThousands)
                fi.NumberGroupSizes = new int[] {3, 0};
            else
                fi.NumberGroupSizes = new int[] {0};
            fi.NumberNegativePattern = 1;
            return Get(unit).ToString("N", fi) + UnitToName(unit);
        }

        static public bool TryParse(string text, out Pressure value, out Unit unit)
        {
            value = null;
            unit = DefaultUnit;

            //split value to text and unit name
            string number = null;
            string name = null;
            for (int i = text.Length - 1; i >= 0; i--)
            {
                if (Char.IsDigit(text[i]))
                {
                    name = text.Substring(i + 1);
                    number = text.Substring(0, i + 1);
                    break;
                }
            }

            if (name == null || name.Length == 0)
                return false;

            try
            {
                unit = NameToUnit(name);
            }
            catch (Exception )
            {
                return false;
            }

            NumberFormatInfo fi = new NumberFormatInfo();
            fi.CurrencyDecimalSeparator = ".";
            fi.NumberGroupSeparator = ",";
            fi.NumberNegativePattern = 1;

            double v;
            if (!Double.TryParse(number, out v))
                return false;

            value = new Pressure(v, unit);
            return true;
        }

        public Pressure ToUnit(Unit unit)
        {
            return new Pressure(Get(unit), unit);
        }
    }
}