﻿using System;
using System.Globalization;

namespace MathEx.ExternalBallistic.Units
{
    public class Temperature
    {
        private double mValue;            //value in Fahrenheit
        static private Temperature mConvertor = new Temperature(0, Unit.Fahrenheit);
        static private object mMutex = new object();

        public enum Unit
        {
            Celsius,
            Fahrenheit,
        }

        public Temperature(double value, Unit unit)
        {
            Set(value, unit);
        }


        private Unit mSetUnit;

        public Unit SetUnit
        {
            get
            {
                return mSetUnit;
            }
        }

        protected void Set(double value, Unit unit)
        {
            mSetUnit = unit;
            switch (unit)
            {
            case    Unit.Fahrenheit:
                    mValue = value;
                    break;
            case    Unit.Celsius:
                    mValue = value * 9 / 5 + 32;
                    break;
            }
        }

        public double Get(Unit unit)
        {
            switch (unit)
            {
            case    Unit.Fahrenheit:
                    return mValue;
            case    Unit.Celsius:
                    return (mValue - 32) * 5 / 9;
            }
            throw new ArgumentException("Unknown unit");
        }

        static public double Convert(double value, Unit from, Unit to)
        {
            lock (mMutex)
            {
                mConvertor.Set(value, from);
                return mConvertor.Get(to);
            }
        }

        static public string UnitToName(Unit unit)
        {
            switch (unit)
            {
            case    Unit.Fahrenheit:
                    return "°F";
            case    Unit.Celsius:
                    return "°C";
            }
            throw new ArgumentException("Unknown unit");
        }

        static public Unit NameToUnit(string name)
        {
            if (name == "°F")
                return Unit.Fahrenheit;
            else if (name == "°C")
                return Unit.Celsius;
            throw new ArgumentException("Unknown unit");
        }

        static public int DefaultDisplayPrecision(Unit unit)
        {
            return 0;
        }

        static public Unit DefaultUnit
        {
            get
            {
                return Unit.Fahrenheit;
            }
        }

        override public string ToString()
        {
            return ToString(DefaultUnit, true, DefaultDisplayPrecision(DefaultUnit));
        }

        public string ToString(Unit unit)
        {
            return ToString(unit, true, DefaultDisplayPrecision(unit));
        }

        public string ToString(Unit unit, bool groupThousands, int precision)
        {
            NumberFormatInfo fi = new NumberFormatInfo();
            fi.NumberDecimalDigits = precision;
            fi.CurrencyDecimalSeparator = ".";
            fi.NumberGroupSeparator = ",";
            if (groupThousands)
                fi.NumberGroupSizes = new int[] {3, 0};
            else
                fi.NumberGroupSizes = new int[] {0};
            fi.NumberNegativePattern = 1;
            return Get(unit).ToString("N", fi) + UnitToName(unit);
        }

        static public bool TryParse(string text, out Temperature value, out Unit unit)
        {
            value = null;
            unit = DefaultUnit;

            //split value to text and unit name
            string number = null;
            string name = null;
            for (int i = text.Length - 1; i >= 0; i--)
            {
                if (Char.IsDigit(text[i]))
                {
                    name = text.Substring(i + 1);
                    number = text.Substring(0, i + 1);
                    break;
                }
            }

            if (name == null || name.Length == 0)
                return false;

            try
            {
                unit = NameToUnit(name);
            }
            catch (Exception )
            {
                return false;
            }

            NumberFormatInfo fi = new NumberFormatInfo();
            fi.CurrencyDecimalSeparator = ".";
            fi.NumberGroupSeparator = ",";
            fi.NumberNegativePattern = 1;

            double v;
            if (!Double.TryParse(number, out v))
                return false;

            value = new Temperature(v, unit);
            return true;
        }

        public Temperature ToUnit(Unit unit)
        {
            return new Temperature(Get(unit), unit);
        }
    }
}