﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.Xml;
using MathEx.ExternalBallistic.Units;

namespace Gehtsoft.BallisticCalculator.Reticle
{
    public static class ReticleController
    {
        public static class Calculation
        {
            public static double getResultion(int cx1, int cy1, int cx2, int cy2, Angle distance)
            {
                double pixels = Math.Sqrt(Math.Pow(cx1 - cx2, 2) + Math.Pow(cy1 - cy2, 2));
                return distance.Get(Angle.Unit.Mil) / pixels;
            }

            public static Angle ImagePixelsToAngle(int p, int z, double resolution, Angle.Unit setUnit)
            {
                int distance = p - z;
                double value = distance * resolution;
                Angle angle = new Angle(value, Angle.Unit.Mil);
                angle = angle.ToUnit(setUnit);
                return angle;
            }

            public static int AngleToPixels(Angle angle, int z, double resolution)
            {
                double value = angle.Get(Angle.Unit.Mil);
                return (int)Math.Round(z + value / resolution);
            }
        }


        public static class Serialization
        {
            public static Reticle loadReticle(string fileName, bool saveImageSource)
            {
                XmlReaderSettings settings;
                settings = new XmlReaderSettings();
                settings.ConformanceLevel = ConformanceLevel.Document;
                settings.IgnoreWhitespace = true;
                settings.IgnoreComments = true;

                XmlDocument doc = new XmlDocument();
                using (XmlReader reader = XmlReader.Create(fileName))
                    doc.Load(reader);

                string name = "Unknown";
                int centerX = 0, centerY = 0, calibrationX1 = 0, calibrationY1 = 0, calibrationX2 = 0, calibrationY2 = 0;
                Angle calibrationAngle = new Angle(1, Angle.Unit.MilDot);

                Image image = null;
                bool imageRaster = true;
                byte[] imageSource = null;

                List<KeyValuePair<int, int>> bdcs = new List<KeyValuePair<int, int>>();

                if (doc.DocumentElement.Name == "reticle")
                {
                    if (doc.DocumentElement.Attributes["name"] != null)
                        name = doc.DocumentElement.Attributes["name"].Value;

                    foreach (XmlNode node in doc.DocumentElement.ChildNodes)
                    {
                        if (node.Name == "center")
                        {
                            //compat with 1.0
                            if (node.Attributes["x"] != null)
                                if (!Int32.TryParse(node.Attributes["x"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out centerX))
                                    centerX = -1;
                            if (node.Attributes["y"] != null)
                                if (!Int32.TryParse(node.Attributes["y"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out centerY))
                                    centerY = -1;
                        }
                        if (node.Name == "zero")
                        {
                            if (node.Attributes["x"] != null)
                                if (!Int32.TryParse(node.Attributes["x"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out centerX))
                                    centerX = -1;
                            if (node.Attributes["y"] != null)
                                if (!Int32.TryParse(node.Attributes["y"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out centerY))
                                    centerY = -1;
                        }
                        if (node.Name == "calibration")
                        {
                            if (node.Attributes["x1"] != null)
                                if (!Int32.TryParse(node.Attributes["x1"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out calibrationX1))
                                    calibrationX1 = -1;
                            if (node.Attributes["y1"] != null)
                                if (!Int32.TryParse(node.Attributes["y1"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out calibrationY1))
                                    calibrationY1 = -1;
                            if (node.Attributes["x2"] != null)
                                if (!Int32.TryParse(node.Attributes["x2"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out calibrationX2))
                                    calibrationX2 = -1;
                            if (node.Attributes["y2"] != null)
                                if (!Int32.TryParse(node.Attributes["y2"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out calibrationY2))
                                    calibrationY2 = -1;
                            Angle ar = null;
                            if (node.Attributes["angle"] != null)
                            {

                                Angle.Unit u;
                                if (!Angle.TryParse(node.Attributes["angle"].Value, out ar, out u))
                                    ar = null;
                            }
                            calibrationAngle = ar;
                        }
                        else if (node.Name == "resolution")
                        {
                            //compat with 1.0
                            int r = -1;
                            if (node.Attributes["pixel"] != null)
                                if (!Int32.TryParse(node.Attributes["pixel"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out r))
                                    r = -1;
                            Angle ar = null;
                            if (node.Attributes["angle"] != null)
                            {

                                Angle.Unit u;
                                if (!Angle.TryParse(node.Attributes["angle"].Value, out ar, out u))
                                    ar = null;
                            }


                            if (ar != null && r >= 0)
                            {
                                calibrationX1 = centerX;
                                calibrationX2 = centerX;
                                calibrationY1 = centerY;
                                calibrationY2 = centerY + r;
                                calibrationAngle = ar;
                            }
                        }
                        else if (node.Name == "bdc")
                        {
                            foreach (XmlNode node1 in node.ChildNodes)
                            {
                                if (node1.Name == "point")
                                {
                                    int xx = -1;
                                    int yy = -1;
                                    if (node1.Attributes["y"] != null)
                                        if (!Int32.TryParse(node1.Attributes["y"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out yy))
                                            yy = -1;
                                    if (node1.Attributes["x"] != null)
                                        if (!Int32.TryParse(node1.Attributes["x"].Value, NumberStyles.Integer, CultureInfo.InvariantCulture, out xx))
                                            xx = -1;

                                    if (yy >= 0)
                                    {
                                        if (xx < 0)
                                            xx = centerX;       //compat with 0.9

                                        bdcs.Add(new KeyValuePair<int, int>(xx, yy));
                                    }
                                }
                            }
                        }
                        else if (node.Name == "image")
                        {
                            XmlAttribute attribute = node.Attributes["type"];

                            if (attribute == null || attribute.Value == null || attribute.Value == "image/png")
                            {
                                string text = node.InnerText;
                                try
                                {
                                    byte[] arr = Convert.FromBase64CharArray(text.ToCharArray(), 0, text.Length);
                                    imageSource = arr;
                                    using (MemoryStream ms = new MemoryStream(arr))
                                        image = new Bitmap(Image.FromStream(ms));
                                }
                                catch (Exception)
                                {
                                    image = null;
                                }
                            }
                            else if (attribute.Value == "windows/metafile")
                            {
                                string text = node.InnerText;
                                try
                                {
                                    byte[] arr = Convert.FromBase64CharArray(text.ToCharArray(), 0, text.Length);
                                    imageSource = arr;
                                    using (MemoryStream ms = new MemoryStream(arr))
                                        image = new Metafile(ms);
                                }
                                catch (Exception)
                                {
                                    image = null;
                                }
                                imageRaster = false;
                            }
                        }
                    }
                }
                Reticle reticle = new Reticle();
                reticle.Name = name;
                reticle.Image = image;
                reticle.Raster = imageRaster;
                reticle.ZeroX = centerX;
                reticle.ZeroY = centerY;
                reticle.CalibrationX1 = calibrationX1;
                reticle.CalibrationY1 = calibrationY1;
                reticle.CalibrationX2 = calibrationX2;
                reticle.CalibrationY2 = calibrationY2;
                reticle.CalibrationAngle = calibrationAngle;
                if (saveImageSource)
                    reticle.ImageSource = imageSource;
                foreach (KeyValuePair<int, int> point in bdcs)
                    reticle.BDCPoints.Add(new Reticle.BDCPoint(point.Key, point.Value));
                return reticle;
            }

            public static void saveReticle(Reticle reticle, string filename)
            {
                if (reticle == null ||
                    reticle.Name == null ||
                    reticle.Image == null ||
                    reticle.CalibrationAngle == null ||
                    reticle.ImageSource == null ||
                    reticle.ZeroX < 0 ||
                    reticle.ZeroY < 0 ||
                    reticle.CalibrationX1 < 0 ||
                    reticle.CalibrationX2 < 0 ||
                    reticle.CalibrationY1 < 0 ||
                    reticle.CalibrationY2 < 0)
                    throw new ArgumentException("Reticle is not completely filled", "reticle");

                

                using (StreamWriter output = new StreamWriter(filename, false, Encoding.UTF8))
                {
                    output.WriteLine("<reticle name=\"{0}\">", reticle.Name);
                    output.WriteLine(" <zero x=\"{0}\" y=\"{1}\"/>", reticle.ZeroX.ToString(CultureInfo.InvariantCulture), reticle.ZeroY.ToString(CultureInfo.InvariantCulture));
                    output.WriteLine(" <calibration x1=\"{0}\" y1=\"{1}\" x2=\"{2}\" y2=\"{3}\" angle=\"{4}\"/>",
                        reticle.CalibrationX1.ToString(CultureInfo.InvariantCulture),
                        reticle.CalibrationY1.ToString(CultureInfo.InvariantCulture),
                        reticle.CalibrationX2.ToString(CultureInfo.InvariantCulture),
                        reticle.CalibrationY2.ToString(CultureInfo.InvariantCulture),
                        reticle.CalibrationAngle.ToString(reticle.CalibrationAngle.SetUnit));
                    if (reticle.BDCPoints.Count > 0)
                    {
                        output.WriteLine(" <bdc>");
                        foreach (Reticle.BDCPoint point in reticle.BDCPoints)
                            output.WriteLine("  <point x=\"{0}\" y=\"{1}\" />", point.X.ToString(CultureInfo.InvariantCulture), point.Y.ToString(CultureInfo.InvariantCulture));
                        output.WriteLine(" </bdc>");
                    }
                    string b64, type;
                    if (reticle.Raster)
                        type = "image/png";
                    else
                        type = "windows/metafile";
                
                    b64 = Convert.ToBase64String(reticle.ImageSource, Base64FormattingOptions.InsertLineBreaks);

                    output.WriteLine(" <image type=\"{0}\">", type);
                    output.WriteLine(b64);
                    output.WriteLine(" </image>");
                    output.WriteLine("</reticle>");
                }
            }
        }
    }
}
