﻿using System;
using System.Collections.Generic;
using System.Text;
using MathEx.ExternalBallistic;
using MathEx.ExternalBallistic.Units;

namespace Gehtsoft.BallisticCalculator.UI
{
    public enum MeasurementSystem
    {
        Metric,
        Imperial
    }

    public enum ColumnData : uint
    {
        Range = 0,
        Velocity = 1,
        Mach = 2,
        Energy = 3,
        Path = 4,
        Hold = 5,
        Windage = 6,
        WindageAdj = 7,
        FlightTime = 8,
        OGW = 9,
        __MAX = 10,
    }
    
    public class BallisticModel
    {
        private BallisticInfoCollection mBallisticInfo = null;

        public BallisticInfoCollection BallisticInfo
        {
            get
            {
                return mBallisticInfo;
            }
            set
            {
                mBallisticInfo = value;
           }
        }

        private MeasurementSystem mMeasurementSystem;

        public MeasurementSystem MeasurementSystem
        {
            get
            {

                return mMeasurementSystem;
            }
            set
            {
                mMeasurementSystem = value;
            }
        }

        private Angle.Unit mAngleUnits = Angle.Unit.MilDot;

        public Angle.Unit AngleUnits
        {
            get
            {
                return mAngleUnits;
            }
            set
            {
                mAngleUnits = value;
            }
        }


        public bool GetOneRowDataDisplay(int item, ref string[] arr)
        {
            return GetOneRowDataDisplay(mBallisticInfo, item, ref arr);
        }

        protected bool GetOneRowDataDisplay(BallisticInfoCollection collection, int item, ref string[] arr)
        {
            if (collection == null || item < 0 || item >= collection.Count)
                return false;

            BallisticInfo info = collection[item];

            if (arr == null || arr.Length != (uint)ColumnData.__MAX)
                arr = new string[(uint)ColumnData.__MAX];

            for (int i = 0; i < arr.Length; i++)
                arr[i] = null;

            if (mMeasurementSystem == UI.MeasurementSystem.Imperial)
            {
                arr[(uint)ColumnData.Range] = info.Range.ToString(Distance.Unit.Yard, true, 0);
                arr[(uint)ColumnData.Velocity] = info.BulletVelocity.ToString(Velocity.Unit.FeetPerSecond, true, 1);
                arr[(uint)ColumnData.Energy] = info.BulletEnergy.ToString(Energy.Unit.FootPounds, true, 0);
                arr[(uint)ColumnData.Path] = info.Path.ToString(Distance.Unit.Inch, true, 2);
                arr[(uint)ColumnData.Windage] = info.Windage.ToString(Distance.Unit.Inch, true, 2);
                arr[(uint)ColumnData.OGW] = info.OptimalGameWeight.ToString(Weight.Unit.Pound, true, 0);

            }
            else if (mMeasurementSystem == UI.MeasurementSystem.Metric)
            {
                arr[(uint)ColumnData.Range] = info.Range.ToString(Distance.Unit.Meter, true, 0);
                arr[(uint)ColumnData.Velocity] = info.BulletVelocity.ToString(Velocity.Unit.MeterPerSecond, true, 1);
                arr[(uint)ColumnData.Energy] = info.BulletEnergy.ToString(Energy.Unit.Joule, true, 0);
                arr[(uint)ColumnData.Path] = info.Path.ToString(Distance.Unit.Centimeter, true, 2);
                arr[(uint)ColumnData.Windage] = info.Windage.ToString(Distance.Unit.Centimeter, true, 2);
                arr[(uint)ColumnData.OGW] = info.OptimalGameWeight.ToString(Weight.Unit.Kilogram, true, 0);
            }
            arr[(uint)ColumnData.Mach] = info.Mach.ToString("f2");
            if (item == 0)
            {
                arr[(uint)ColumnData.Hold] = "";
                arr[(uint)ColumnData.WindageAdj] = "";
            }
            else
            {
                arr[(uint)ColumnData.Hold] = info.Hold.ToString(mAngleUnits, false, 2);
                arr[(uint)ColumnData.WindageAdj] = info.WindageCorrection.ToString(mAngleUnits, false, 2);

            }
            arr[(uint)ColumnData.FlightTime] = ConvertFlightTime(info.Time); 
            return true;
        }

        public void GetHeaderDataExport(ref string[] arr)
        {
            GetHeaderDataExport(mBallisticInfo, ref arr);
        }

        protected void GetHeaderDataExport(BallisticInfoCollection collection, ref string[] arr)
        {
            if (arr == null || arr.Length != (uint)ColumnData.__MAX)
                arr = new string[(uint)ColumnData.__MAX];

            for (int i = 0; i < arr.Length; i++)
                arr[i] = null;

            if (mMeasurementSystem == UI.MeasurementSystem.Imperial)
            {
                arr[(uint)ColumnData.Range] = "Range (yard)";
                arr[(uint)ColumnData.Velocity] = "Velocity (fps)";
                arr[(uint)ColumnData.Energy] = "Energy (ft-lbs)";
                arr[(uint)ColumnData.Path] = "Path (in)";
                arr[(uint)ColumnData.Windage] = "Windage (in)";
                arr[(uint)ColumnData.OGW] = "Optimal Game Weight (lbs)";
            }
            else if (mMeasurementSystem == UI.MeasurementSystem.Metric)
            {
                arr[(uint)ColumnData.Range] = "Range (m)";
                arr[(uint)ColumnData.Velocity] = "Velocity (m/s)";
                arr[(uint)ColumnData.Energy] = "Energy (J)";
                arr[(uint)ColumnData.Path] = "Path (cm)";
                arr[(uint)ColumnData.Windage] = "Windage (cm)";
                arr[(uint)ColumnData.OGW] = "Optimal Game Weight (kg)";

            }

            string au = Angle.UnitToName(mAngleUnits);

            arr[(uint)ColumnData.Mach] = "Mach";
            arr[(uint)ColumnData.Hold] = "Hold (" + au + ");";
            arr[(uint)ColumnData.WindageAdj] = "Windage Adjustment (" + au + ");";
            arr[(uint)ColumnData.FlightTime] = "Time (sec)";
        }

        public bool GetOneRowDataExport(int item, ref string[] arr)
        {
            return GetOneRowDataExport(mBallisticInfo, item, ref arr);
        }

        protected bool GetOneRowDataExport(BallisticInfoCollection collection, int item, ref string[] arr)
        {
            if (collection == null || item < 0 || item >= collection.Count)
                return false;

            BallisticInfo info = collection[item];

            if (arr == null || arr.Length != (uint)ColumnData.__MAX)
                arr = new string[(uint)ColumnData.__MAX];

            for (int i = 0; i < arr.Length; i++)
                arr[i] = null;

            double range = 0, velocity = 0, energy = 0, path = 0, windage = 0, ogw = 0;

            if (mMeasurementSystem == UI.MeasurementSystem.Imperial)
            {
                range = info.Range.Get(Distance.Unit.Yard);
                velocity = info.BulletVelocity.Get(Velocity.Unit.FeetPerSecond);
                energy = info.BulletEnergy.Get(Energy.Unit.FootPounds);
                path = info.Path.Get(Distance.Unit.Inch);
                windage = info.Windage.Get(Distance.Unit.Inch);
                ogw = info.OptimalGameWeight.Get(Weight.Unit.Pound);

            }
            else if (mMeasurementSystem == UI.MeasurementSystem.Metric)
            {
                range = info.Range.Get(Distance.Unit.Meter);
                velocity = info.BulletVelocity.Get(Velocity.Unit.MeterPerSecond);
                energy = info.BulletEnergy.Get(Energy.Unit.Joule);
                path = info.Path.Get(Distance.Unit.Centimeter);
                windage = info.Windage.Get(Distance.Unit.Centimeter);
                ogw = info.OptimalGameWeight.Get(Weight.Unit.Kilogram);
            }
            arr[(uint)ColumnData.Range] = range.ToString("f2");
            arr[(uint)ColumnData.Velocity] = velocity.ToString("f2");
            arr[(uint)ColumnData.Energy] = energy.ToString("f2");
            arr[(uint)ColumnData.Path] = path.ToString("f2");
            arr[(uint)ColumnData.Windage] = windage.ToString("f2");
            arr[(uint)ColumnData.OGW] = ogw.ToString("f2");
            arr[(uint)ColumnData.Mach] = info.Mach.ToString("f2");
            if (item == 0)
            {
                arr[(uint)ColumnData.Hold] = "";
                arr[(uint)ColumnData.WindageAdj] = "";
            }
            else
            {
                arr[(uint)ColumnData.Hold] = info.Hold.Get(mAngleUnits).ToString("f2");
                arr[(uint)ColumnData.WindageAdj] = info.WindageCorrection.Get(mAngleUnits).ToString("f2");
            }
            arr[(uint)ColumnData.FlightTime] = ConvertFlightTime(info.Time); 
            return true;
        }

        protected string ConvertFlightTime(TimeSpan ts)
        {
            string s = ts.ToString();
            int l = 9;
            if (s.Length < 12)
                l = s.Length - 3;
            return (s.Substring(3, l));
        }

        private ColumnData mGraphData = ColumnData.Path;

        public ColumnData GraphData
        {
            get
            {
                return mGraphData;
            }
            set
            {
                mGraphData = value;
            }
        }

        public class GraphPair
        {
            private double mX, mY;
            private string mTag;

            public double X
            {
                get
                {
                    return mX;
                }
            }

            public double Y
            {
                get
                {
                    return mY;
                }
            }

            public string Tag
            {
                get
                {
                    return mTag;
                }
            }

            public GraphPair(double x, double y, string tag)
            {
                mX = x;
                mY = y;
                mTag = tag;
            }
        }

        public List<GraphPair> GetGraphData()
        {
            return GetGraphData(mBallisticInfo);
        }

        public string GetGraphName()
        {
            switch (mGraphData)
            {
                case ColumnData.Path:
                    return "Path";
                case ColumnData.Hold:
                    return "Hold";
                case ColumnData.Mach:
                    return "Mach";
                case ColumnData.Windage:
                    return "Windage";
                case ColumnData.WindageAdj:
                    return "Windage Adjustment";
                case ColumnData.OGW:
                    return "Optimal Game Weight";
                case ColumnData.Energy:
                    return "Energy";
                case ColumnData.Velocity:
                    return "Velocity"; 
                case ColumnData.FlightTime:
                    return "Flight Time";
                default:
                    return "Unknown";
            }
        }

        protected List<GraphPair> GetGraphData(BallisticInfoCollection collection)
        {
            if (collection == null)
                return null;

            List<GraphPair> pairs = new List<GraphPair>();

            for (int i = 0; i < collection.Count; i++)
            {
                double x, y;
                string tag, at, value;

                x = collection[i].Range.Get(MeasurementSystem == UI.MeasurementSystem.Imperial ? Distance.Unit.Yard : Distance.Unit.Meter);
                at = collection[i].Range.ToString(MeasurementSystem == UI.MeasurementSystem.Imperial ? Distance.Unit.Yard : Distance.Unit.Meter, true, 0);

                switch (mGraphData)
                {
                    case ColumnData.Path:
                        y = collection[i].Path.Get(MeasurementSystem == UI.MeasurementSystem.Imperial ? Distance.Unit.Inch : Distance.Unit.Centimeter);
                        value = collection[i].Path.ToString(MeasurementSystem == UI.MeasurementSystem.Imperial ? Distance.Unit.Inch : Distance.Unit.Centimeter, true, 1);
                        break;
                    case ColumnData.Hold:
                        if (i == 0)
                            continue;
                        y = collection[i].Hold.Get(AngleUnits);
                        value = collection[i].Hold.ToString(AngleUnits, false, 1);
                        break;
                    case ColumnData.Mach:
                        y = collection[i].Mach;
                        value = y.ToString("f2");
                        break;
                    case ColumnData.Windage:
                        y = collection[i].Windage.Get(MeasurementSystem == UI.MeasurementSystem.Imperial ? Distance.Unit.Inch : Distance.Unit.Centimeter);
                        value = collection[i].Windage.ToString(MeasurementSystem == UI.MeasurementSystem.Imperial ? Distance.Unit.Inch : Distance.Unit.Centimeter, true, 1);
                        break;
                    case ColumnData.WindageAdj:
                        if (i == 0)
                            continue;
                        y = collection[i].WindageCorrection.Get(AngleUnits);
                        value = collection[i].WindageCorrection.ToString(AngleUnits, false, 1);
                        break;
                    case ColumnData.OGW:
                        y = collection[i].OptimalGameWeight.Get(MeasurementSystem == UI.MeasurementSystem.Imperial ? Weight.Unit.Pound : Weight.Unit.Kilogram);
                        value = collection[i].OptimalGameWeight.ToString(MeasurementSystem == UI.MeasurementSystem.Imperial ? Weight.Unit.Pound : Weight.Unit.Kilogram, false, 0);
                        break;
                    case ColumnData.Energy:
                        y = collection[i].BulletEnergy.Get(MeasurementSystem == UI.MeasurementSystem.Imperial ? Energy.Unit.FootPounds : Energy.Unit.Joule);
                        value = collection[i].BulletEnergy.ToString(MeasurementSystem == UI.MeasurementSystem.Imperial ? Energy.Unit.FootPounds : Energy.Unit.Joule, true, 0);
                        break;
                    case ColumnData.Velocity:
                        y = collection[i].BulletVelocity.Get(MeasurementSystem == UI.MeasurementSystem.Imperial ? Velocity.Unit.FeetPerSecond : Velocity.Unit.MeterPerSecond);
                        value = collection[i].BulletVelocity.ToString(MeasurementSystem == UI.MeasurementSystem.Imperial ? Velocity.Unit.FeetPerSecond : Velocity.Unit.MeterPerSecond, true, 1);
                        break;
                    case ColumnData.FlightTime:
                        y = collection[i].Time.TotalSeconds;
                        value = ConvertFlightTime(collection[i].Time);
                        break;
                    default:
                        return null;
                }

                tag = string.Format("{0} @ {1}", value, at);
                pairs.Add(new GraphPair(x, y, tag));
            }
            return pairs;
        }
    }

    public class BallisticInfoMultiModel : BallisticModel
    {
        private BallisticInfoCollection[] mSeries = null;

        public BallisticInfoMultiModel(int seriesCount)
        {
            mSeries = new BallisticInfoCollection[seriesCount];
        }

        public void SetSeria(int seria, BallisticInfoCollection collection)
        {
            if (seria >= 0 && seria < mSeries.Length)
                mSeries[seria] = collection;
            if (seria == 0)
                BallisticInfo = collection;
        }

        public bool HasSeria(int seria)
        {
            if (seria < 0 || seria >= mSeries.Length)
                return false;
            return mSeries[seria] != null;
        }

        public void RemoveSeria(int seria)
        {
            if (HasSeria(seria))
                mSeries[seria] = null;
            if (seria == 0)
                BallisticInfo = null;
        }

        public string SeriaName(int seria)
        {
            if (seria < 0 || seria >= mSeries.Length)
                return "";
            return mSeries[seria].Name;

        }

        public List<GraphPair> GetGraphData(int seria)
        {
            if (!HasSeria(seria))
                return null;

            return GetGraphData(mSeries[seria]);
        }

    }
}
